
/*
  Kjetil Matheussen, 2005/2006.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sndfile.h>
#include <pthread.h>
#include <math.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>



#define JC_MAX(a,b) (((a)>(b))?(a):(b))
#define JC_MIN(a,b) (((a)<(b))?(a):(b))

#define OPTARGS_BEGIN(das_usage) {int lokke;const char *usage=das_usage;for(lokke=1;lokke<argc;lokke++){char *a=argv[lokke];if(!strcmp("--help",a)||!strcmp("-h",a)){printf(usage);return 0;
#define OPTARG(name,name2) }}else if(!strcmp(name,a)||!strcmp(name2,a)){{
#define OPTARG_GETINT() atoi(argv[++lokke])
#define OPTARG_GETFLOAT() atof(argv[++lokke])
#define OPTARG_GETSTRING() argv[++lokke]
#define OPTARG_LAST() }}else if(lokke==argc-1){lokke--;{
#define OPTARGS_ELSE() }else if(1){
#define OPTARGS_END }else{fprintf(stderr,usage);return(-1);}}}



/* Arguments and their default values */
static float buffer_time=60;
static jack_client_t *client=NULL;
#define DEFAULT_NUM_CHANNELS 2
static int channels=-1;
static int bitdepth=0;
static char *base_filename=NULL;
static char *filename=NULL;
static int leading_zeros=1;
static double recording_time=0.0;
static int disk_error_stop=0;
static int disk_errors=0;
static char *soundfile_format="wav";
static bool silent=false;

/* JACK data */
static jack_port_t **ports;
typedef jack_default_audio_sample_t sample_t;
static int jack_buffer_size;
static int jack_buffer_size_is_changed_to=0;
static float jack_samplerate;
static int connect_num=0;
static int unreported_overruns=0;


/* Disk thread */
static pthread_t disk_thread={0};
static pthread_mutex_t disk_thread_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  data_ready = PTHREAD_COND_INITIALIZER;


/* Synchronization between jack process thread and disk thread. */
static volatile int is_initialized=0; // This $@#$@#$ variable is needed because jack ports must be initialized _after_ (???) the client is activated. (stupid jack)
static volatile int is_running=1;
static jack_ringbuffer_t *rb=NULL;
struct ringbuffer_block{
  int overruns;
  sample_t *buffer;
};


/* Jack connecting thread. */
static pthread_t connect_thread={0} ;
static pthread_cond_t  connect_ready = PTHREAD_COND_INITIALIZER;





/////////////////////////////////////////////////////////////////////
//////////////////////// BUFFERS ////////////////////////////////////
/////////////////////////////////////////////////////////////////////


static sample_t **buffers=NULL;
static sample_t *empty_buffer=NULL;

static int num_buffers;
static int buffer_size;
static int buffer_size_in_bytes;

static void buffers_init(int jackbuffersize){

  if(jack_buffer_size!=jackbuffersize)
    if(silent==false)
      printf("(Reallocating buffers to fit jack buffer size of %d)\n",jackbuffersize);

  // Samplebuffers;
  {
    static sample_t *das_buffer=NULL;
    int lokke;
    
    free(das_buffer);
    free(buffers);
    free(empty_buffer);
    
    buffer_size=channels*jackbuffersize;
    buffer_size_in_bytes=buffer_size*sizeof(sample_t);

    num_buffers=2 + (
		     buffer_time*jack_samplerate*sizeof(sample_t)*channels 
		     / buffer_size_in_bytes);
    

    // Allocate Ringbuffer
    {
      if(rb!=NULL)
	jack_ringbuffer_free(rb);
      
      rb=jack_ringbuffer_create(sizeof(struct ringbuffer_block)*num_buffers);
      //printf("ringbuffer_size: %d, num_buffers: %d\n",rb->size/4,num_buffers);
      
      /* When JACK is running realtime, jack_activate() will have
       * called mlockall() to lock our pages into memory.  But, we
       * still need to touch any newly allocated pages before
       * process() starts using them.  Otherwise, a page fault could
       * create a delay that would force JACK to shut us down. (from jackrec, -Kjetil)*/
      memset(rb->buf,0,rb->size);
    }

    //The size of the ringbuffer must be less than num_buffers, but jack_ringbuffer_create allocates
    //up to the nearest 2^n number. Therefore we need to set num_buffers again.
    num_buffers=rb->size/sizeof(struct ringbuffer_block);
    num_buffers+=3;


    //printf("buffer_size: %d, buffer_size_in_bytes: %d, num_buffers: %d, total: %d, buffer_time: %f\n",
    //       buffer_size,buffer_size_in_bytes,num_buffers,num_buffers*buffer_size_in_bytes,buffer_time);
    
    das_buffer=calloc(num_buffers,buffer_size_in_bytes);
    buffers=calloc(sizeof(sample_t*),num_buffers);
    
    for(lokke=0;lokke<num_buffers;lokke++){
      buffers[lokke]=das_buffer+(lokke*buffer_size);
    }
    
    empty_buffer=calloc(1,buffer_size_in_bytes);
  }

  
  {
    jack_buffer_size=jackbuffersize;
    jack_buffer_size_is_changed_to=0;
  }
}


static sample_t *buffer_get(void){
  static int next_free_buffer=0;
  sample_t *ret=buffers[next_free_buffer];
  next_free_buffer++;
  if(next_free_buffer==num_buffers)
    next_free_buffer=0;
  return ret;
}

int buffersizecallback(size_t newsize,void *arg){  
  jack_buffer_size_is_changed_to=newsize;
  if(pthread_mutex_trylock(&disk_thread_lock)==0){
    pthread_cond_signal(&data_ready);
    pthread_mutex_unlock(&disk_thread_lock);
  }

  return 0;
}



/////////////////////////////////////////////////////////////////////
//////////////////////// PORTNAMES //////////////////////////////////
/////////////////////////////////////////////////////////////////////
static char **cportnames=NULL;
static int num_cportnames=0; // After initialization, this variable is not used anymore and "channels" is used instead.
                             // This is because "channels" can be set on the commandline.

static int findnumports(char **ports){
  int ret=0;
  while(ports && ports[ret]!=NULL)
    ret++;
  return ret;
}


static void portnames_add_defaults(void){
  if(cportnames==NULL){
    cportnames=(char **)jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
    num_cportnames=JC_MIN(DEFAULT_NUM_CHANNELS,findnumports(cportnames));
    if(num_cportnames==0){
      printf("No physical output ports found in your jack setup. Exiting.\n");
      exit(0);
    }
  }
  if(channels==-1){
    channels=num_cportnames;
  }else{
    num_cportnames=JC_MIN(channels,num_cportnames);
    channels=num_cportnames;
  }
  if(channels==0){
    printf("No point recording 0 channels. Exiting.\n");
    exit(0);
  }
}

static void portnames_add(char *name){
  char **outportnames=(char**)jack_get_ports(client,name,"",0);
  int ch=findnumports(outportnames);

  if(ch>0){
    cportnames=realloc(cportnames,(num_cportnames+ch)*sizeof(char*));

    ch=0;
    while(outportnames[ch]!=NULL){
      cportnames[num_cportnames]=outportnames[ch];
      ch++;
      num_cportnames++;
    }

  }else{
    printf("\nWarning, No ports with name \"%s\".\n",name);
    if(cportnames==NULL)
      if(silent==false)
	printf("This could lead to using default ports instead.\n");
  }
}


static char **portnames_get_connections(int ch){
  jack_port_t* port=jack_port_by_name(client,cportnames[ch]);
  char **ret;

  if(jack_port_flags(port) & JackPortIsInput){
    ret=(char**)jack_port_get_all_connections(client,port);
  }else{
    ret=calloc(2,sizeof(char*));
    ret[0]=cportnames[ch];
  }
  
  return ret;
}





/////////////////////////////////////////////////////////////////////
//////////////////////// DISK ///////////////////////////////////////
/////////////////////////////////////////////////////////////////////

// These four variables are used in case we break the 4GB barriere for standard wav files.
static int num_files=1;
static int64_t disksize=0;
static bool is_using_wav=true;
static int bytes_per_frame;

static SNDFILE *soundfile;
static unsigned long long overruns=0;



#include "setformat.c"


static void open_soundfile(void){
  int wav_type;

  SF_INFO sf_info={0};
  sf_info.samplerate = jack_get_sample_rate (client);
  sf_info.channels = channels;
  
  setformat(&sf_info,soundfile_format);
  if(sf_info.format==SF_FORMAT_WAV)
    is_using_wav=true;
  else
    is_using_wav=false;

  bytes_per_frame=bitdepth/8;

  switch (bitdepth) {
  case 8: wav_type = SF_FORMAT_PCM_U8;
    break;
  case 16: wav_type = SF_FORMAT_PCM_16;
    break;
  case 24: wav_type = SF_FORMAT_PCM_24;
    break;
  case 32: wav_type = SF_FORMAT_PCM_32;
    break;
  default:
    wav_type = 
      (!strcasecmp("flac",soundfile_format)) ? SF_FORMAT_PCM_24 :
      SF_FORMAT_FLOAT;
    bytes_per_frame=4;
    break;
  }

  sf_info.format |= wav_type;

  if(filename==NULL)
    filename=strdup(base_filename);

  soundfile=sf_open(filename,SFM_WRITE,&sf_info);
  if(soundfile==NULL){
    fprintf (stderr, "\ncannot open sndfile \"%s\" for output (%s)\n", filename,sf_strerror(NULL));
    jack_client_close (client);
    exit(1);
  }
}


static void close_soundfile(void){
  sf_close (soundfile);
  if (overruns > 0) {
    fprintf (stderr,
	     "\njack_capture failed with a total of %llu overruns.\n", overruns);
    fprintf (stderr, " try a bigger buffer than -B %f\n",buffer_time);
  }
  if (disk_errors > 0)
    fprintf(stderr,"\nWarning: jack_capture failed with a total of %d disk errors.\n",disk_errors);
}


static void handle_filelimit(size_t frames){
  int new_bytes=frames*bytes_per_frame*channels;

  if(is_using_wav && (disksize + new_bytes >= UINT32_MAX-(1024*1024))){ // (1024*1024) should be enough for the header.
    sf_close(soundfile);{
      char *filename_new;
      filename_new=calloc(1,strlen(base_filename)+500);      
      sprintf(filename_new,"%s.%0*d.wav",base_filename,2,num_files);
      printf("Warning. 4GB limit on wav file almost reached. Closing %s, and continue writing to %s.\n",filename,filename_new);
      num_files++;
      free(filename);
      filename=filename_new;
    }open_soundfile();
    disksize=0;
  }

  disksize+=new_bytes;
}


static int disk_write(void *buffer,size_t frames){
  handle_filelimit(frames);
  
  if(sf_writef_float(soundfile,buffer,frames) != frames){
    fprintf (stderr,
	     "Error. Cannot write sndfile (%s)\n",
	     sf_strerror(soundfile));
    disk_errors++;
    return 0;
  }
  return 1;
}


static int disk_write_overruns(int num_overruns){
  int lokke;
  fprintf (stderr,
	   "\nWarning. jack_capture failed with %d overrun%s. Some parts of the recording will contain silence.\n"
	   "try a bigger buffer than -B %f\n"
	   "Continue recording...\n",
	   num_overruns,num_overruns==1?"":"s",
	   buffer_time);
  overruns+=num_overruns;
  for(lokke=0;lokke<num_overruns;lokke++)
    if( ! disk_write(empty_buffer,jack_buffer_size))
      return 0;
  return 1;
}


static void *disk_thread_func (void *arg){
       open_soundfile();

	// Main disk loop
	{
	  for(;;){

	    while(is_initialized && (jack_ringbuffer_read_space (rb) >= sizeof(struct ringbuffer_block))){
	      struct ringbuffer_block block;
	      sample_t *buffer=NULL;
	      
	      jack_ringbuffer_read(rb,(void*)&block,sizeof(struct ringbuffer_block));
	      buffer=block.buffer;
	      
	      if(block.overruns>0 &&
		 ( ! disk_write_overruns(block.overruns))
		 && disk_error_stop)
		goto done;
	      
	      if( (! disk_write(buffer,jack_buffer_size) )
		  && disk_error_stop)
		goto done;
	    }
	      
	    if(is_running==0)
	      goto done;
	    
	    if(jack_buffer_size_is_changed_to>0)
	      buffers_init(jack_buffer_size_is_changed_to);
	    
	    /* wait until process() signals more data */
	    pthread_cond_wait(&data_ready, &disk_thread_lock);
	  }
	}


 done:

	if(unreported_overruns>0)
	  disk_write_overruns(unreported_overruns);

	close_soundfile();

	{
	  pthread_mutex_unlock(&disk_thread_lock);
	  if(silent==false)
	    printf ("disk thread finished\n");
	}

	return 0;
}

void setup_disk_thread (void){
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  pthread_mutex_lock(&disk_thread_lock);
  pthread_create(&disk_thread, NULL, disk_thread_func, NULL);
}


void stop_disk_thread(void){
 /* Wake up disk thread. (no trylock) (isrunning==0)*/
  pthread_cond_signal(&data_ready);
  pthread_join(disk_thread, NULL);
}





/////////////////////////////////////////////////////////////////////
//////////////////////// JACK ///////////////////////////////////////
/////////////////////////////////////////////////////////////////////	
void put_jack_buffers_into_buffer(sample_t *buffer){
  jack_default_audio_sample_t *in[channels];
  int ch;
  int i,pos=0;
  
  for(ch=0;ch<channels;ch++)
    in[ch]=jack_port_get_buffer(ports[ch],jack_buffer_size);
      
  for(i=0;i<jack_buffer_size;i++)
    for(ch=0;ch<channels;ch++)
      buffer[pos++]=in[ch][i];
}

void send_buffer_to_disk_thread(sample_t *buffer){
  struct ringbuffer_block block;

  block.overruns=unreported_overruns;
  block.buffer=buffer;
  jack_ringbuffer_write(rb,(void*)&block,sizeof(struct ringbuffer_block));
  unreported_overruns=0;
  
  if (pthread_mutex_trylock (&disk_thread_lock) == 0) {
    pthread_cond_signal (&data_ready);
    pthread_mutex_unlock (&disk_thread_lock);
  }
}

int process (jack_nframes_t nframes, void *arg){

  if(is_initialized==0)
    return 0;

  if((  jack_buffer_size_is_changed_to != 0)  // The inserted silence will have the wrong size, but at least the user will get a report that something was not recorded.
     || jack_ringbuffer_write_space(rb)<sizeof(struct ringbuffer_block)
     )
    {
      unreported_overruns++;
    }else{
      sample_t *buffer=buffer_get(); // This can't fail because the size of the ringbuffer is less than the number of buffers.
      put_jack_buffers_into_buffer(buffer);
      send_buffer_to_disk_thread(buffer);
    }
  
  return 0;
}






/////////////////////////////////////////////////////////////////////
/////////////////// JACK CONNECTIONS ////////////////////////////////
/////////////////////////////////////////////////////////////////////


static pthread_mutex_t connect_thread_lock = PTHREAD_MUTEX_INITIALIZER;


static int compare(const void *a, const void *b){
  return strcmp((const char*)a,(const char*)b);
}

static int reconnect_ports_questionmark(void){
  int ch;
  for(ch=0;ch<channels;ch++){
    char **connections1 = portnames_get_connections(ch);
    const char **connections2 = jack_port_get_all_connections(client,ports[ch]);
    int memb1=findnumports(connections1);
    int memb2=findnumports((char**)connections2);

    if(memb1!=memb2){
      free(connections1);
      free(connections2);
      return 1;
    }
    
    qsort(connections1,memb1,sizeof(char*),compare);
    qsort(connections2,memb1,sizeof(char*),compare);
    
    {
      int lokke = 0;
      for(lokke=0;lokke<memb1;lokke++){
	//printf("connect_ports \"%s\" \"%s\" \n",connections1[lokke],connections2[lokke]);
	if(strcmp(connections1[lokke],connections2[lokke])){
	  free(connections1);
	  free(connections2);
	  return 1;
	}
      }
    }

    free(connections1);
    free(connections2);
  }

  return 0;
}



static void disconnect_ports(void){
  int ch;
  for(ch=0;ch<channels;ch++){
    int lokke = 0;
    const char **connections = jack_port_get_all_connections(client,ports[ch]);
    while(connections && connections[lokke] != NULL){
      jack_disconnect(client,connections[lokke],jack_port_name(ports[ch]));
      lokke++;
    }
    free(connections);
  }
}


static void connect_ports(void){
  int ch;
  for(ch=0;ch<channels;ch++){
    int lokke = 0;
    char **connections = portnames_get_connections(ch);
    while(connections && connections[lokke] != NULL){
      int err=jack_connect(client,connections[lokke],jack_port_name(ports[ch]));
      if(err!=0)
	fprintf(stderr, "\ncannot connect input port %s to %s, errorcode %s\n", jack_port_name (ports[ch]), connections[lokke],strerror(err));      
      lokke++;
    }
    free(connections);
  }
}

static bool connection_thread_started=false;

static void* connection_thread(void *arg){
  connection_thread_started=true;
  while(1){
    pthread_cond_wait(&connect_ready, &connect_thread_lock);
    if(is_running==0)
      goto done;
    if(is_initialized && reconnect_ports_questionmark()){
      if(silent==false)
	printf("\n(Reconnecting ports)\n");
      disconnect_ports();
      connect_ports();
    }
  }

 done:
  pthread_mutex_unlock(&connect_thread_lock);
  if(silent==false)
    printf("connection thread finished\n");
  return NULL;
}

static void wake_up_connection_thread(void){
  //Don't want trylock here.
  while(connection_thread_started==false) usleep(10);
  pthread_cond_signal (&connect_ready);
}

static void start_connection_thread(void){
  pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  pthread_mutex_lock (&connect_thread_lock);
  pthread_create(&connect_thread,NULL,connection_thread,NULL);
}

static void stop_connection_thread(void){
  wake_up_connection_thread();
  pthread_join(connect_thread, NULL);
}

static int graphordercallback(void *arg){
  wake_up_connection_thread();
  return 0;
}



static void create_ports(void){
  ports = (jack_port_t **) calloc (sizeof (jack_port_t *),channels);  
  {
    int ch;
    for(ch=0;ch<channels;ch++) {
      char name[500];
      sprintf(name,"input%d",ch+1);
      ports[ch]=jack_port_register(client,name,JACK_DEFAULT_AUDIO_TYPE,JackPortIsInput,0);
      if(ports[ch]==0){
	fprintf(stderr,"\ncannot register input port \"%s\"!\n", name);
	jack_client_close(client);
	exit(1);
      }
    }
  }
}






/////////////////////////////////////////////////////////////////////
/////////////////// INIT / SHUTDOWN /////////////////////////////////
/////////////////////////////////////////////////////////////////////

static void do_exit(int close_jack,int returnval){
  static bool is_called=false;

  if(is_called==false){
    is_called=true;
    is_running=0;

    stop_disk_thread();
    
    if(close_jack) // Don't close if called from jack_shutdown. (its already closed)
      jack_client_close(client);
    
    stop_connection_thread();

    exit(returnval);
  }
}


static pid_t mainpid;
static void finish(int sig){
  if(getpid()==mainpid){
    if(silent==false)
      printf("I'm signaled to finish. Please wait while writing all data to disk.\n");
    do_exit(1,0);
  }
}

static void jack_shutdown(void *arg){
  fprintf (stderr, "\nJACK shutdown.\n");
  do_exit(0,0);
}


static jack_client_t *new_jack_client(char *name){
  jack_client_t *client=jack_client_new(name);

  if(client==NULL){
    for(connect_num=1;connect_num<100;connect_num++){
      char temp[500];
      sprintf(temp,"%s_%d",name,connect_num);
      client=jack_client_new(temp);
      if(client!=NULL) return client;
    }
    fprintf(stderr, "\njack server not running? (Unable to create jack client \"%s\")\n",name);
    exit(1);
  }
  return client;
}


static void start_jack(void){
  static bool I_am_already_called=false;
  if(I_am_already_called) // start_jack is called more than once if the --port argument has been used.
    return;

  client=new_jack_client("jack_capture");

  jack_buffer_size=jack_get_buffer_size(client);
  jack_samplerate=jack_get_sample_rate(client);
  I_am_already_called=true;
}


int main (int argc, char *argv[]){
  mainpid=getpid();
  // Arguments
  {
    OPTARGS_BEGIN("Usage: jack_capture  [--bitdepth n] [--bufsize seconds] [--channels n] [--port port] [filename]\n"
		  "                     [ -b n]        [ -B seconds]       [ -c n]        [ -p port]\n"
		  "\n"
		  "\"filename\" is by default auotogenerated to something like \"jack_capture_<number>.wav\"\n"
		  "\"bitdepth\" is by default FLOAT. It can be set to either 8, 16, 24 or 32.\n"
		  "\"channels\" is by default 2.\n"
		  "\"bufsize\"  is by default 60 seconds.\n"
		  "\"port\"     is by default set to all ports connected to the physical outputs. Can be specified\n"
		  "           more than once.\n"
		  "\n"
		  "\n"
		  "Additional arguments:\n"
		  "[--duration s] or [-d s]         -> Recording is stopped after \"s\" seconds.\n"
		  "[--leading-zeros n] or [-z n]    -> \"n\" is the number of zeros to in the autogenerated filename.\n"
		  "                                    (-z 2 -> jack_capture_001.wav, and so on.) (default is 1)\n"
		  "[--format format] or [-f format] -> See http://www.mega-nerd.com/libsndfile/api.html#open\n"
		  "                                    (Default is wav)\n"
		  "[--version] or [-v]              -> Prints out version.\n"
		  "[--silent] or [-s]               -> Suppress some common messages printed to the terminal.\n"
		  "\n"
		  "Examples:\n"
		  "\n"
		  "To record a stereo file of what you hear:\n"
		  "  $jack_capture\n"
		  "\n"
		  "To record a stereo file of what you hear in the flac format:\n"
		  " $jack_capture -f flac\n"
		  "\n"
		  "To record a stereo file of what you hear (same as default if using alsa):\n"
		  "  $jack_capture --port alsa_pcm:playback_1 --port alsa_pcm:playback_2\n"
		  "\n"
		  "Same as above, but shorter syntax:\n"
		  "  $jack_capture --channels 2 --port alsa_pcm:playback*\n"
		  "\n"
		  "To record the output from jamin:\n"
		  "  $jack_capture --port jamin:out* sound_from_jamin.wav\n"
		  "\n"
		  "To record all sound coming in to jamin:\n"
		  "  $jack_capture --port jamin:in* sound_to_jamin.wav\n"
		  "\n"
		  "To record all sound coming in and out of jamin:\n"
		  "  $jack_capture --port jamin* sound_to_and_from_jamin.wav\n"
		  "\n"
		  )
      {
	OPTARG("--bitdepth","-b") bitdepth = OPTARG_GETINT();
	OPTARG("--bufsize","-B") buffer_time = OPTARG_GETFLOAT();
	OPTARG("--channels","-c") channels = OPTARG_GETINT();
	OPTARG("--leading-zeros","-z") leading_zeros = OPTARG_GETINT();
	OPTARG("--recording-time","-d") recording_time = OPTARG_GETFLOAT();
	OPTARG("--port","-p") start_jack() ; portnames_add(OPTARG_GETSTRING());
	OPTARG("--format","-f") soundfile_format=OPTARG_GETSTRING();
	OPTARG("--version","-v") puts(VERSION);exit(0);
	OPTARG("--silent","-s") silent=true;
	OPTARG_LAST() base_filename=OPTARG_GETSTRING();
      }OPTARGS_END;
  }
  

  // Filename
  {
    if(base_filename==NULL){
      int try=0;
      base_filename=calloc(1,5000);
      for(;;){
	sprintf(base_filename,"jack_capture_%0*d.%s",leading_zeros+1,++try,soundfile_format);
	if(access(base_filename,F_OK)) break;
      }
    }
  }
  

  // Init jack 1
  {
    start_jack();
    portnames_add_defaults();
  }


  //
  {
    buffers_init(jack_buffer_size);
  }


  //
  {
    setup_disk_thread ();
  }


  // Init jack 2
  {
    jack_set_process_callback(client, process, NULL);
    jack_on_shutdown(client, jack_shutdown, NULL);
    
    jack_set_graph_order_callback(client,graphordercallback,NULL);
    jack_set_buffer_size_callback(client,buffersizecallback,NULL);

    if (jack_activate(client)) {
      fprintf (stderr,"\ncannot activate client");
    }
   
    create_ports();
    connect_ports();
    start_connection_thread();
  }


  // Everything initialized.
  //   (The threads are waiting for this variable, not the other way around, so now it just needs to be set.)
  {
    is_initialized=1;
  }


  {
    signal(SIGINT,finish);

    
    if(recording_time>0.0){
      printf("Recording to \"%s\". The recording is going to last %lf seconds. Press <Ctrl-C> to stop before that.\n",base_filename,recording_time);  
      sleep(recording_time);
      usleep( (recording_time-floor(recording_time)) * 1000000);
    }else{
      // Wait for <return> or SIGINT
      printf("Recording to \"%s\". Press <Return> or <Ctrl-C> to stop it.\n",base_filename);        
      {
	char gakk[64];
	fgets(gakk,49,stdin);
      }
      if(silent==false)
	printf("Please wait while writing all data to disk. (shouldn't take long)\n");
    }
  }


  //
  {
    do_exit(1,0);
  }

  
  return 0;
}


